"""
Storage Utilities module
"""
from datetime import datetime, timedelta
import itertools
from typing import Iterator
from azure.storage.blob import BlobServiceClient, ContainerClient, ResourceTypes, AccountSasPermissions, generate_account_sas


def generate_azure_account_sas(connection_string: str, valid_for_hours: int = 24) -> str:
    """
    Generate SAS token for provided connection string

    Args:
        connection_string (str): connection string
        valid_for_hours (int, optional): SAS-token validity in hours. Defaults to 24.

    Returns:
        str: SAS-token
    """
    blob_service_client = BlobServiceClient.from_connection_string(connection_string)

    sas_token = generate_account_sas(
        blob_service_client.account_name,
        account_key=blob_service_client.credential.account_key,
        resource_types=ResourceTypes(object=True, container=True),
        permission=AccountSasPermissions(read=True, write=True, list=True, create=True),
        expiry=datetime.utcnow() + timedelta(hours=valid_for_hours)
    )

    return sas_token


def walk_storage_files(container_uri: str, container_path: str, page_size: int) -> Iterator[str]:
    """
    Performes storage traversal, calls new_item_lambda for every item

    Args:
        container_uri (str): Blob storage container uri
        container_path (str): Path in the container
        page_size (int): size of the page for list_blobs operation

    Yields:
        Iterator[str]: iterator over blob names, lazily following list_blobs continuation tokens
    """
    container_client = ContainerClient.from_container_url(
        container_url=container_uri)

    for blob in itertools.islice(
            container_client.list_blobs(name_starts_with=container_path, results_per_page=page_size),  # noqa: E127
            page_size):  # noqa: E127
        if blob.name.endswith("/"):
            continue
        yield blob.name



# pylint: disable=unnecessary-pass,too-few-public-methods
"""
CSV-to-file static mapper:
maps file paths taken from a CVS column to a data files in a remote storage
"""
import os
import glob
import csv
import re
from typing import List
from .static_file_mapper import StaticFileMapper


class CSVToFileMapper(StaticFileMapper):
    """
    CSV to File Mapper

    Attributes
    ----------
    csv_column : str
        CSV column index to get file paths from (starts from 0)
    skip_header : bool
        skip first line in CSV (usually column header)
    labels_pattern : str
        pattern for using with glob.iglob when getting CSV files

    Methods
    -------
    map_files():
        Performs file mapping.
    """
    def __init__(self, cache_dir: str, labels_dir: str, labels_pattern: str, csv_column: int = 0, skip_header: bool = True):
        """
        Initializes file mapper

        Args:
            cache_dir (str): data cache directory
            labels_dir (str): labels directory
            labels_pattern (str): labels pattern
            csv_column (int, optional): CVS column index for getting file paths from. Defaults to 0.
            skip_header (bool, optional): Whether to skip CVS header (first line). Defaults to True.
        """
        super().__init__(cache_dir, labels_dir)
        self.csv_column = csv_column
        self.skip_header = skip_header
        self.labels_pattern = labels_pattern

    def map_files(self) -> List[str]:
        """
        Performes file mapping. Derived classes must override it.

        Returns:
            List[str]: list of mapped files
        """
        paths = []
        last_file = ""  # we preserve last file name as a simple way to avoid duplicates
        for filepath in glob.iglob(os.path.join(self.labels_dir, self.labels_pattern), recursive=True):
            with open(filepath, "r") as csvfile:
                datareader = csv.reader(csvfile)
                try:
                    if self.skip_header:
                        next(datareader)  # skip the header row
                    for row in datareader:
                        file = str(row[self.csv_column]).strip()

                        # removing filename "tails" with and after '#' and '?'
                        # (e.g. timestamps in VoTT, filename.mp4#timestamp)
                        file = re.split('#|\\?', file)[0]

                        if file not in ("", last_file):
                            paths.append(file)
                            last_file = file
                except StopIteration:
                    pass
        return paths



"""
MLDL module entry point
"""
import sys
from .cli.cli_main import main as _main


def main():
    """module entry point function
    """
    _main(sys.argv[1:])


if __name__ == '__main__':
    main()



# pylint: disable=bad-option-value,too-many-branches,too-complex,too-many-locals,too-many-statements
"""
Data repo CLI module
"""
import argparse
import datetime
import os
import shutil
import tempfile
from distutils import dir_util
from typing import List

import yaml

from ..datarepo.mappers.csv_to_file import CSVToFileMapper
from ..datarepo.mappers.file_to_file import FileToFileMapper
from ..datarepo.storage.azure_blob import AzureBlobStorage
from ..datarepo.storage.storage_utils import generate_azure_account_sas

_CONFIG_FILE_DEFAULT_NAME = "datarepo.yml"
_DEFAULT_ERROR = 255


def _cli_main(parser: argparse.ArgumentParser, args_list: List[str]):
    """
    Runs MLDL CLI commands

    Args:
        parser (argparse.ArgumentParser): argument parser object
        args_list (List[str]): command line arguments
    """
    # datarepo.yaml contains a specific data repository configuration configuration
    # Expected configuration variables and example values:
    # data_azure_storage_container_uri: "https://storageaccountname.blob.core.windows.net/container_name"
    #   - data storage container uri
    # data_azure_storage_path: "folder/subfolder" - data storage path (under the container)
    # data_azure_storage_connection_string: "..."
    #       - connection string for the data storage account or a SAS-token 
    #         (SAS token should start with '?' or 'sas:')
    #
    # datarepo.yaml must be either in current directory or in .local directory next to this script

    args = parser.parse_args(args_list)

    do_list = False
    do_cache = False
    do_link = False
    with_storage = False
    use_temp_cache = False

    labels_dir = os.path.abspath(os.path.normpath(args.labels_dir))
    if not os.path.exists(labels_dir):
        print(f"Labels directory \"{labels_dir}\" does not exist")
        parser.exit(_DEFAULT_ERROR)

    if args.cache_dir is not None:
        cache_dir = os.path.abspath(os.path.normpath(args.cache_dir))
    else:
        cache_dir = None

    if "cache" in args.stage or "list" in args.stage:
        with_storage = True

    if with_storage:
        # Check config file
        if args.config_file is None:
            config_file = _CONFIG_FILE_DEFAULT_NAME
            if not os.path.exists(config_file):
                config_file = f".local/{_CONFIG_FILE_DEFAULT_NAME}"
                if not os.path.exists(config_file):
                    print(f"Cannot find {_CONFIG_FILE_DEFAULT_NAME}. "
                            "It must be either in the current directory "  # noqa: E127
                            "or in .local directory or "  # noqa: E127
                            "at a path provided with --config option.")  # noqa: E127
                    parser.exit(_DEFAULT_ERROR)
        else:
            config_file = args.config_file
            if not os.path.exists(config_file):
                print(f"{config_file} doesn't exist or not accessible.")
                parser.exit(_DEFAULT_ERROR)

        # parse config file
        with open(config_file) as cfg_file:
            config = yaml.safe_load(cfg_file)

        if "list" in args.stage:
            do_list = True
            if cache_dir is None:
                cache_dir = tempfile.mkdtemp()
                use_temp_cache = True
        else:
            if cache_dir is None:
                print("--cache value is required for caching.")
                parser.exit(_DEFAULT_ERROR)
            do_cache = True

        # configure storage
        os.makedirs(cache_dir, exist_ok=True)
        storage = AzureBlobStorage(cache_dir,
                                   config['data_azure_storage_container_uri'],
                                   config['data_azure_storage_path'] 
                                   if 'data_azure_storage_path' in config
                                        else None)

        storage_connection = ""
        if 'data_azure_storage_connection_string' in config:
            storage_connection = config['data_azure_storage_connection_string']
        elif args.storage_connection is not None:
            storage_connection = args.storage_connection

        if storage_connection is None or storage_connection == "":
            print("Storage connection string or a SAS-token must be provided in the configuration file "  # noqa: E127
                "or as --storage-connection option.")
            parser.exit(_DEFAULT_ERROR)

        # if storage_connection is an SAS-token, starts with '?'
        if storage_connection.startswith('?'):
            storage.authenticate(storage_connection[1:])
        # if storage_connection is an SAS-token, starts with 'sas:'
        elif storage_connection.casefold().startswith('sas:'):
            storage.authenticate(storage_connection[4:])
        # otherwise it's a connection string, and we generate an SAS token before authenticating
        else:
            storage.authenticate(generate_azure_account_sas(storage_connection))

    if not do_list and "link" in args.stage:
        if args.data_dir is None:
            print("--data value is required for linking")
            parser.exit(_DEFAULT_ERROR)

        do_link = True

    if not do_list:
        if args.labels_template is None:
            print("--labels_template value is required for file mapper")
            parser.exit(_DEFAULT_ERROR)
        elif args.labels_template.startswith("/"):
            args.labels_template = args.labels_template[1:]
        else:
            args.labels_template = f"**/*{args.labels_template}"

        if args.mapper_type == "file":
            if args.data_ext is None:
                print("--data_extension value is required for file mapper")
                parser.exit(_DEFAULT_ERROR)

            mapper = FileToFileMapper(
                cache_dir, labels_dir, args.labels_template, args.data_ext)
        elif args.mapper_type == "csv":
            mapper = CSVToFileMapper(
                cache_dir, labels_dir, args.labels_template, int(args.csv_col), not args.csv_use_header)

        files = mapper.map_files()
        total_len = len(files)
        print(f"Mapped {total_len} file entries.")
    else:
        # performing listing
        print(f"Listing storage in groups of {args.group_len} items.")
        storage.generate_storage_lists(int(args.group_len))
        print("Copying listing files")
        dir_util.copy_tree(cache_dir, labels_dir, preserve_mode=0, preserve_times=0)
        print(f"Done listing. Files are in {labels_dir}")

    if do_cache:
        print(f"Caching files to {cache_dir}.")
        start_time = datetime.datetime.now()
        storage.cache_items(files)
        end_time = datetime.datetime.now()
        print(f"Caching completed. Total time: {(end_time-start_time).total_seconds()} seconds.")

    if do_link:
        data_dir = os.path.abspath(os.path.normpath(args.data_dir))
        good_files = 0
        print(f"Caching files from {data_dir} to {cache_dir}.")

        for file in files:
            repo_file_path = os.path.abspath(os.path.join(data_dir, file))
            cached_path = storage.get_cached_path(file)
            if os.path.exists(cached_path):
                if os.path.exists(repo_file_path):
                    if os.path.islink(repo_file_path):
                        os.unlink(repo_file_path)
                    else:
                        os.remove(repo_file_path)
                dir_name = os.path.dirname(repo_file_path)
                if dir_name:
                    os.makedirs(dir_name, exist_ok=True)

                repo_file_dir = os.path.dirname(repo_file_path)
                cached_relative = os.path.relpath(cached_path, repo_file_dir)
                os.symlink(cached_relative, repo_file_path)
                good_files += 1
            else:
                print(f"File {repo_file_path} was not cached. Skipping.")

        print(f"Linking completed. {good_files} new links have been made.")

    if use_temp_cache:
        shutil.rmtree(cache_dir)


def _create_parser() -> argparse.ArgumentParser:
    """
    Creates command line parser for MLDL tool

    Returns:
        argparse.ArgumentParser: argument parser
    """
    parser = argparse.ArgumentParser(
        description='Machine Learning Data Lineage Tool')

    parser.add_argument('-s', '--stage', help='Stage.',
                        nargs='+', action='store', dest='stage', required=True,
                        choices=['list', 'cache', 'link'])

    parser.add_argument('-m', '--mapper', help='Labels to data mapper type. Required for cache and map stages.',
                        action='store', dest='mapper_type', required=False)

    parser.add_argument('--labels_template',
                        help="Labels file name template (e.g. \".json\" or \"/name.csv\"). "
                        "Forward slash means specific name. Without it, template resolves with '**/*' prefix.",
                        action='store', dest='labels_template', required=False)

    parser.add_argument('--data_extension', help='Data file extension (e.g. \".jpg\"). Only required for file mapper.',
                        action='store', dest='data_ext', required=False)

    parser.add_argument('--csv_column_index', help="Index of CSV column with file names. "
                        "First column index is 0 (default). Only used with csv mapper.",
                        action='store', dest='csv_col', default=0, required=False)

    parser.add_argument('--csv_no_header', help='Specify if CSV files do not have header line. Only used with csv mapper.',
                        action='store_false', dest='csv_use_header', default=False, required=False)

    parser.add_argument('-c', '--cache', help='Cache directory.',
                        action='store', dest='cache_dir', required=False)

    parser.add_argument('-l', '--labels', help='Labels directory.',
                        action='store', dest='labels_dir', required=True)

    parser.add_argument('-d', '--data', help='Data directory. Required for link stage.',
                        action='store', dest='data_dir', required=False)

    parser.add_argument('--group_len', help='Maximum number of files in a listing file. Required for list stage.',
                        action='store', dest='group_len', default=1000)

    parser.add_argument('--config', help='Configuration file. Required for cache and list stages.',
                        action='store', dest='config_file', required=False)
    
    parser.add_argument('--storage-connection', 
                        help='Storage connection string or SAS token. Used with cache and list stages when is not in the configuration file.',
                        action='store', dest='storage_connection', required=False)

    return parser


def main(args: List[str]):
    """
    Main function

    Args:
        args (List[str]): command line arguments
    """
    parser = _create_parser()
    _cli_main(parser, args)



# pylint: disable=unnecessary-pass,too-few-public-methods
"""
Base class for static file mappers
"""
from typing import List


class StaticFileMapper:
    """
    File to File Mapper

    Attributes
    ----------
    cache_dir : str
        Data cache directory
    labels_dir : str
        Labels directory

    Methods
    -------
    map_files():
        Performes file mapping. Derived classes must override it.
    """
    def __init__(self, cache_dir: str, labels_dir: str):
        """
        Initializes file mapper

        Args:
            cache_dir (str): [description]
            labels_dir (str): [description]
        """
        self.cache_dir = cache_dir
        self.labels_dir = labels_dir

    def map_files(self) -> List[str]:
        """
        Performes file mapping. Derived classes must override it.

        Returns:
            List[str]: list of mapped files
        """
        pass



# pylint: disable-all
# flake8: noqa
"""
Unit tests for mappers
"""
import pytest
from unittest.mock import Mock, patch, MagicMock, mock_open
import glob
import builtins
from ..mappers.csv_to_file import CSVToFileMapper
from ..mappers.file_to_file import FileToFileMapper


def _fake_iglob(pathname, *, recursive=False):
    return ['1.csv']


def _fake_rel_path(path, start=None):
    return path


def test_csv_mapper():
    my_open = mock_open(read_data='filename\n00000000.jpg\n')
    with patch('builtins.open', my_open, create=True):
        with patch('glob.iglob', _fake_iglob, create=True):
            mapper = CSVToFileMapper("cache", "labels", ".csv")
            files = mapper.map_files()
            assert len(files) == 1
            assert files[0] == "00000000.jpg"


def test_file_mapper():
    with patch('glob.iglob', _fake_iglob, create=True):
        with patch('os.path.relpath', _fake_rel_path, create=True):
            mapper = FileToFileMapper("cache", "labels", ".csv", ".mp4")
            files = mapper.map_files()
            assert len(files) == 1
            assert files[0] == "1.mp4"



# pylint: disable=unnecessary-pass,too-few-public-methods
"""
Base class for remote storage providers
"""
from typing import List


class Storage:
    """
    Base class for storage providers.
    """
    def __init__(self, cache_dir: str, storage_container_uri: str, storage_path: str = None):
        """
        Initializes storage provider

        Args:
            cache_dir (str): data cache directory
            storage_container_uri (str): remote storage container uri
            storage_path (str, optional): remote storage path under container. Defaults to None.
        """
        self.cache_dir = cache_dir
        self.storage_container_uri = storage_container_uri
        self.storage_container_name = storage_container_uri.rsplit('/', 1)[-1]
        self.storage_path = storage_path

    def authenticate(self, sas_token: str):
        """
        Uses provided security token for authentication on the remote storage

        Args:
            sas_token (str): security token
        """
        pass

    def is_cached(self, path: str) -> bool:
        """
        Checks if provided item is in the data cache

        Args:
            path (str): item to check

        Returns:
            bool: True if provided item is in the data cache
        """
        pass

    def get_cached_path(self, path: str) -> str:
        """
        returns full path to a locally cached item

        Args:
            path (str): item to return local path for

        Returns:
            str: full local path
        """
        pass

    def cache_items(self, paths: List[str]) -> bool:
        """
        Performes caching operation for provided items

        Args:
            paths (List[str]): List of items to cache

        Returns:
            bool: True if operation was successful
        """
        pass

    def generate_storage_lists(self, group_len: int):
        """
        Generates lists with all items in the remote storage
        and saves them to files with group_len items per file.
        Files are saved to cache_dir.

        Args:
            group_len (int): maximum number of items per file
        """
        pass



# pylint: disable-all
# flake8: noqa
"""
Unit tests for main CLI module
"""
from unittest.mock import Mock, patch, MagicMock, mock_open
import glob
import builtins
from functools import partial
from pathlib import Path
from ...datarepo.storage.azure_blob import AzureBlobStorage
from azure.storage.blob import ContainerClient
from ..cli_main import main


_TEST_CONTAINER_URL = "https://fake.blob.core.windows.net/containername"
_TEST_PATH_NAME = "folder/subfolder"

_original_open = builtins.open


class FakeBlob:
    def __init__(self, name):
        self.name = name


def _mock_list_blobs(*args, **kwargs):
    blobs = [
        FakeBlob('folder/subfolder/file1.mp4'),
        FakeBlob('folder/subfolder/subfolder2/'),
        FakeBlob('folder/subfolder/subfolder2/file2.mp4')
    ]

    return iter(blobs)


def _write_func(text):
    assert text == "filename\n" or text == "file1.mp4\n" or text == "subfolder2/file2.mp4\n"

def _get_fake_config():
    fake_text = f"data_azure_storage_container_uri: \"{_TEST_CONTAINER_URL}\"\n"
    fake_text += f"data_azure_storage_path: \"{_TEST_PATH_NAME}\"\n"
    fake_text += "data_azure_storage_connection_string: \"DefaultEndpointsProtocol=https;" \
                "AccountName=fake;AccountKey=123==;EndpointSuffix=core.windows.net\""
    return fake_text

def _mocked_file(m, fn, *args, **kwargs):
    m.opened_file = Path(fn)
    if fn.endswith(".csv"):
        file_obj = mock_open(read_data="").return_value
        file_obj.write = _write_func
        return file_obj
    elif fn.endswith("datarepo.yml"):
        confix_text = _get_fake_config()
        return mock_open(read_data=confix_text).return_value
    else:
        return _original_open(fn, *args, **kwargs)

def _create_mock_open():
    mock = mock_open(read_data='')
    mock.side_effect = partial(_mocked_file, mock)
    return mock

def _always_exists(_):
    return True

def test_cli_parameters_success():
    try:
        my_open = _create_mock_open()

        with patch('azure.storage.blob.ContainerClient.list_blobs', _mock_list_blobs):
            with patch('builtins.open', my_open, create=True):
                with patch('os.path.exists', _always_exists, create=True):
                    main(["-s", "list", "-l", "dataset/labels/listed", "--group_len", "10"])
        
        assert True
    except:
        assert False


def test_cli_parameters_fail():
    try:
        main(["-s", "list", "-l", "dataset/labels/listed", "--group_len", "10"])
        assert False
    except SystemExit as ex:
        assert ex.code == 255



# pylint: disable=unnecessary-pass,too-few-public-methods
"""
File-to-file static mapper:
maps a label/annotation file to a data file in a remote storage
with the same name and relative location
"""
import os
import glob
from typing import List
from .static_file_mapper import StaticFileMapper


class FileToFileMapper(StaticFileMapper):
    """
    File to File Mapper

    Attributes
    ----------
    labels_pattern : str
        pattern for using with glob.iglob when getting CSV files
    data_extension : str
        data files extension

    Methods
    -------
    map_files():
        Performs file mapping.
    """

    def __init__(self, cache_dir: str, labels_dir: str, labels_pattern: str, data_extension: str):
        """
        Initializes file mapper

        Args:
            cache_dir (str): data cache directory
            labels_dir (str): labels directory
            labels_pattern (str): labels pattern
            data_extension (str): data files extension
        """
        super().__init__(cache_dir, labels_dir)
        self.labels_pattern = labels_pattern
        self.data_extension = data_extension

        if not self.data_extension.startswith("."):
            self.data_extension = f".{self.data_extension}"

    def map_files(self) -> List[str]:
        """
        Performes file mapping. Derived classes must override it.

        Returns:
            List[str]: list of mapped files
        """
        paths = []

        for file in glob.iglob(os.path.join(self.labels_dir, self.labels_pattern), recursive=True):
            pre, _ = os.path.splitext(file)
            file = pre + self.data_extension

            rel_path = os.path.relpath(file, self.labels_dir)
            paths.append(rel_path)
        return paths



# pylint: disable=bad-option-value,missing-return-doc
"""
A storage provider for working with files Azure Blob Storage
"""
import os
import uuid
from typing import List
import subprocess
from .storage import Storage
from .storage_utils import walk_storage_files


class AzureBlobStorage(Storage):
    """
    Storage provider implementation class for working with files Azure Blob Storage
    """
    _MAX_LIST_GROUP_LEN = 5000
    _PROCESS_TIMEOUT = 96 * 3600

    def __init__(self, cache_dir: str, storage_container_uri: str, storage_path: str = None):
        """
        Initializes storage provider

        Args:
            cache_dir (str): data cache directory
            storage_container_uri (str): remote storage container uri
            storage_path (str, optional): remote storage path under container. Defaults to None.
        """
        self.sas_token = ""
        super().__init__(cache_dir, storage_container_uri, storage_path)

    def authenticate(self, sas_token: str):
        """
        Uses provided security token for authentication on the remote storage

        Args:
            sas_token (str): SAS-token to use with storage operations
        """
        if sas_token is not None:
            if sas_token.startswith("?"):
                self.sas_token = sas_token
            else:
                self.sas_token = f"?{sas_token}"
        else:
            self.sas_token = ""

    def is_cached(self, path: str) -> bool:
        """
        Checks if provided item is in the data cache

        Args:
            path (str): item to check

        Returns:
            bool: True if provided item is in the data cache
        """
        return os.path.exists(self.get_cached_path(path))

    def get_cached_path(self, path: str) -> str:
        """
        returns full path to a locally cached item

        Args:
            path (str): item to return local path for

        Returns:
            str: full local path
        """
        if self.storage_path is not None:
            return os.path.join(self.cache_dir, self.storage_container_name, self.storage_path, path)
        else:
            return os.path.join(self.cache_dir, self.storage_container_name, path)

    def cache_items(self, paths: List[str]) -> bool:
        """
        Performes caching operation for provided items

        Args:
            paths (List[str]): List of items to cache

        Returns:
            bool: True if operation was successful
        """
        return self._az_copy(paths) == 0

    def generate_storage_lists(self, group_len: int):
        """
        Generates lists with all items in the remote storage
        and saves them to files with group_len items per file.
        Files are saved to cache_dir.

        Args:
            group_len (int): maximum number of items per file
        """
        container_uri = self.storage_container_uri
        if self.sas_token is not None and self.sas_token != "":
            container_uri += self.sas_token

        stg_path = self.storage_path
        if stg_path is not None and not stg_path.endswith("/"):
            stg_path += "/"

        current_group = []
        group_number = 0

        for blob in walk_storage_files(container_uri, stg_path, AzureBlobStorage._MAX_LIST_GROUP_LEN):
            if stg_path is not None:
                blob = blob[len(stg_path):]
            current_group.append(blob)
            if len(current_group) == group_len:
                self._store_list(group_number, current_group)
                current_group = []
                group_number += 1

        if len(current_group) != 0:
            self._store_list(group_number, current_group)

    def _store_list(self, list_number: int, file_list: List[str]):
        # 8-digit file names, padded with zeros if needed (00000001.csv)
        file_path = f"{self.cache_dir}/{list_number:08}.csv"

        with open(file_path, 'w') as file_obj:
            file_obj.write("filename\n")
            for item in file_list:
                file_obj.write(f"{item}\n")

    def _get_item_uri(self, local_path: str) -> str:
        if self.storage_path is not None:
            uri = f"{self.storage_container_uri}/{self.storage_path}/{local_path}"
        else:
            uri = f"{self.storage_container_uri}/{local_path}"
        if self.sas_token is not None and self.sas_token != "":
            uri += self.sas_token
        return uri

    def _az_copy(self, paths: List[str], cwd=None) -> int:
        tmp_list_name = f"{uuid.uuid1()}.list.txt"
        tmp_list_path = self.get_cached_path(tmp_list_name)  # temporary file in the cache
        dir_name = os.path.dirname(tmp_list_path)
        if dir_name:
            os.makedirs(dir_name, exist_ok=True)
            with open(tmp_list_path, "w") as tmp_list_file:
                for path in paths:
                    if self.storage_path is not None:
                        tmp_list_file.write(f"{self.storage_path}/{path}\n")
                    else:
                        tmp_list_file.write(f"{path}\n")

        container_uri = self.storage_container_uri
        if self.sas_token is not None and self.sas_token != "":
            container_uri += self.sas_token
        destination = self.cache_dir

        command = f"azcopy cp \"{container_uri}\" \"{destination}\" --list-of-files \"{tmp_list_path}\" "\
                    "--overwrite ifSourceNewer --log-level ERROR"  # noqa: E127
        process = subprocess.Popen(command, cwd=cwd, shell=True)
        try:
            process.wait()
            code = process.poll()
        except subprocess.TimeoutExpired:
            process.kill()
            code = 1

        os.remove(tmp_list_path)

        return code



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mldl',
 'mldl.cli',
 'mldl.cli.tests',
 'mldl.datarepo',
 'mldl.datarepo.mappers',
 'mldl.datarepo.storage',
 'mldl.datarepo.tests']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3,<6.0', 'azure-storage-blob>=12.5,<13.0']

entry_points = \
{'console_scripts': ['mldl = mldl.__main__:main']}

setup_kwargs = {
    'name': 'mldl',
    'version': '0.0.6',
    'description': 'Machine Learning Data Lineage tool (MLDL)',
    'long_description': None,
    'author': 'Vlad Kolesnikov',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# pylint: disable-all
# flake8: noqa
"""
Unit tests for Azure Blob Storage utils (storage_utils.py)
"""
import pytest
from unittest.mock import Mock, patch, MagicMock
from ..storage.storage_utils import walk_storage_files, generate_azure_account_sas
from azure.storage.blob import ContainerClient

_TEST_CONN_STR = "DefaultEndpointsProtocol=https;AccountName=fake;AccountKey=23423423423423423423==;"\
                "EndpointSuffix=core.windows.net"
_TEST_CONTAINER_URL = "https://fake.blob.core.windows.net/containername"
_TEST_PATH_NAME = "folder/subfolder"

def test_sas_success():
    sas_token = generate_azure_account_sas(
        _TEST_CONN_STR,
        1)
    assert sas_token is not None and sas_token != ""


def test_sas_fail():
    try:
        sas_token = generate_azure_account_sas(
            "AccountKey=23423423423423423423==",
            1)
        assert False
    except ValueError:
        assert True
    except:
        assert False

class FakeBlob:
    def __init__(self, name):
        self.name = name

def _mock_list_blobs(*args, **kwargs):
    blobs = [
        FakeBlob('folder/subfolder/file1.mp4'),
        FakeBlob('folder/subfolder/subfolder2/'),
        FakeBlob('folder/subfolder/subfolder2/file2.mp4')
    ]

    return iter(blobs)


def test_walk_storage_files_success():
    with patch('azure.storage.blob.ContainerClient.list_blobs', _mock_list_blobs):
        files_iter = walk_storage_files(
            container_uri=_TEST_CONTAINER_URL,
            container_path=_TEST_PATH_NAME,
            page_size=100)
        files = list(files_iter)
        assert len(files) == 2
        assert files[0] == 'folder/subfolder/file1.mp4'
        assert files[1] == 'folder/subfolder/subfolder2/file2.mp4'



# pylint: disable-all
# flake8: noqa
"""
Unit tests for Azure Blob Storage (azure_blob.py)
"""
import pytest
from unittest.mock import Mock, patch, MagicMock, mock_open
from typing import Callable
from functools import partial
from pathlib import Path
import builtins
from ..storage.azure_blob import AzureBlobStorage
from azure.storage.blob import ContainerClient

_TEST_CONTAINER_URL = "https://fake.blob.core.windows.net/containername"
_TEST_PATH_NAME = "folder/subfolder"

_original_open = builtins.open

class FakeBlob:
    def __init__(self, name):
        self.name = name


def _mock_list_blobs(*args, **kwargs):
    blobs = [
        FakeBlob('folder/subfolder/file1.mp4'),
        FakeBlob('folder/subfolder/subfolder2/'),
        FakeBlob('folder/subfolder/subfolder2/file2.mp4')
    ]

    return iter(blobs)


def _write_func(text):
    assert text == "filename\n" or text == "file1.mp4\n" or text == "subfolder2/file2.mp4\n"

def _mocked_file(m, fn, *args, **kwargs):
    m.opened_file = Path(fn)
    if not fn.endswith(".csv"):
        return _original_open(fn, *args, **kwargs)
    file_obj = mock_open(read_data="").return_value
    file_obj.write = _write_func
    return file_obj


def _create_mock_open():
    mock = mock_open(read_data='')
    mock.side_effect = partial(_mocked_file, mock)
    return mock


def test_generate_storage_lists():
    blob_storage = AzureBlobStorage(
        "cache", _TEST_CONTAINER_URL, _TEST_PATH_NAME)
    
    my_open = _create_mock_open()

    with patch('azure.storage.blob.ContainerClient.list_blobs', _mock_list_blobs):
        with patch('builtins.open', my_open, create=True):
            blob_storage.generate_storage_lists(10)
